unit main;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ComCtrls, ExtCtrls,  Math, ExtDlgs, Jpeg;

type
  PRGB24 = ^TRGB24;
  TRGB24 = record B, G, R: Byte; end;
  PRGBArray = ^TRGBArray;
  TRGBArray = array[0..0] of TRGB24;

  TfrmMain = class(TForm)
    imgDest: TImage;
    imgSrc: TImage;
    pnlControl: TPanel;
    tbAngle: TTrackBar;
    btnOpen: TButton;
    dlgOpen: TOpenPictureDialog;
    radFast: TRadioButton;
    radQuality: TRadioButton;
    chkGrid: TCheckBox;
    dlgSave: TSavePictureDialog;
    btnSave: TButton;
    labAngle: TLabel;
    procedure DoStraighten;
    procedure btnOpenClick(Sender: TObject);
    procedure tbAngleChange(Sender: TObject);
    procedure MakePreviewImage;
    procedure FormCreate(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure chkGridClick(Sender: TObject);
    procedure btnSaveClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  frmMain: TfrmMain;
  mx,my: array[Word]of Integer;
  img: TBitmap;
  imgLoaded: Boolean;

implementation

{$R *.dfm}

procedure SmoothRotate(Src: TBitmap; Angle: Double);
var
  fx, fy, a, tsin, tcos, cxSrc, cySrc, cxDest, cyDest: Double;
  i, ix, dw, dh, width, height, x, y: integer;
  x1, x2, y1, y2: Integer;
  j, k, r, g, b: integer;
  cx, cy, m0, m1: Double;
  C, p0, p1, p2, p3: TRGB24;
  sDst, sDstOff: Integer;
  dest: TBitmap;
  sScanLine: Array of PRGBArray;
begin
  width := Src.Width;
  height := Src.Height;
  a := -(Angle) * pi / 180;
  dw := Round(abs((width) * cos(a)) + abs((height) * sin(a)));
  dh := Round(abs((width) * sin(a)) + abs((height) * cos(a)));
  dest := tBitmap.Create;
  dest.PixelFormat := pf24bit;
  dest.Width := dw;
  dest.Height := dh;
  SetLength(sScanLine, Src.Height);         // ScanLine buffer for Source
  sDst:= Integer(Src.ScanLine[0]);
  sDstOff:= Integer(Src.ScanLine[1]) - Integer(sDst);
  for i:= 0 to Src.Height - 1 do begin
    sScanLine[i]:= PRGBArray(sDst);
    sDst:= sDst + sDstOff;
  end;
  tsin := sin(a);
  tcos := cos(a);
  cxSrc := (src.Width - 1) / 2;
  cySrc := (src.Height - 1) / 2;
  cxDest := (dest.Width - 1) / 2;
  cyDest := (dest.Height - 1) / 2;
  x1 := Trunc(abs((width) * cos(a)) + abs((0) * sin(a))) - 2;
  y1 := Trunc(abs((width) * sin(a)) + abs((0) * cos(a))) + 2;
  x2 := Trunc(abs((0) * cos(a)) + abs((height) * sin(a))) + 2;
  y2 := Trunc(abs((0) * sin(a)) + abs((height) * cos(a))) - 2;
  Dest.Width:= (x1 - x2) + 1;
  Dest.Height:= (y2 - y1) + 1;
  sDst:= Integer(Dest.ScanLine[0]);         // ScanLine for Destiantion
  sDstOff:= Integer(Dest.ScanLine[1]) - sDst;
  for y := y1 to y2 do
  begin
    for x := x2 to x1 do
    begin
      fx := cxSrc + (x - cxDest) * tcos - (y - cyDest) * tsin;
      fy := cySrc + (x - cxDest) * tsin + (y - cyDest) * tcos;
      j := Trunc(fx);
      k := Trunc(fy);
      cx := fx - floor(fx);
      cy := fy - floor(fy);
      p0 := sScanLine[k][j];
      p1 := sScanLine[k][j + 1];
      p2 := sScanLine[k + 1][j];
      p3 := sScanLine[k][j + 1];
      m0 := p0.r + cx * (p1.r - p0.r);
      m1 := p2.r + cx * (p3.r - p2.r) - m0;
      c.R := trunc(m0 + cy * m1);
      m0 := p0.g + cx * (p1.g - p0.g);
      m1 := p2.g + cx * (p3.g - p2.g) - m0;
      c.G := trunc(m0 + cy * m1);
      m0 := p0.b + cx * (p1.b - p0.b);
      m1 := p2.b + cx * (p3.b - p2.b) - m0;
      c.B := trunc(m0 + cy * m1);
      PRGBArray(sDst)[x - x2]:= c;
    end;
    sDst:= sDst + SDstOff;
  end;
  src.Assign(dest);
  dest.free;
  SetLength(sScanLine, 0);
end;

procedure FastRotate(Src: TBitmap; Angle: Double);
var
  w, h, x, y: Integer;
  a: Double;
  i, x1, y1, x2, y2: Integer;
  fx, fy, tsin, tcos, cxSrc, cySrc, cxDest, cyDest: Integer;
  dest: TBitmap;
  sDst, sDstOff: Integer;
  sScanLine: Array of PRGBArray;
begin
  dest:= TBitmap.Create;                    // Make a temp bitmap
  dest.PixelFormat := pf24bit;
  w:= src.Width;
  h:= src.Height;
  a := -(angle) * pi / 180;                 // Calc new image size
  dest.Width := Round(abs((w) * cos(a)) + abs((h) * sin(a)));
  dest.Height:= Round(abs((w) * sin(a)) + abs((h) * cos(a)));
  SetLength(sScanLine, src.Height);         // ScanLine buffer for Source
  sDst:= Integer(src.ScanLine[0]);
  sDstOff:= Integer(src.ScanLine[1]) - Integer(sDst);
  for i:= 0 to Src.Height - 1 do begin
    sScanLine[i]:= PRGBArray(sDst);
    sDst:= sDst + sDstOff;
  end;
  tsin := Round(sin(a) * $FFFF);
  tcos := Round(cos(a) * $FFFF);
  cxSrc := (src.Width - 1) shr 1;
  cySrc := (src.Height - 1) shr 1;
  cxDest := (dest.Width - 1) shr 1;
  cyDest := (dest.Height - 1) shr 1;
  x1 := Trunc(abs((w) * cos(a)) + abs((0) * sin(a))) - 2;
  y1 := Trunc(abs((w) * sin(a)) + abs((0) * cos(a))) + 2;
  x2 := Trunc(abs((0) * cos(a)) + abs((h) * sin(a))) + 2;
  y2 := Trunc(abs((0) * sin(a)) + abs((h) * cos(a))) - 2;
  Dest.Width:= (x1 - x2) + 1;
  Dest.Height:= (y2 - y1) + 1;
  sDst:= Integer(Dest.ScanLine[0]);         // ScanLine for Destiantion
  sDstOff:= Integer(Dest.ScanLine[1]) - sDst;
  for y := y1 to y2 do begin                // Do the rotation
    for x := x2 to x1 do begin
      fx := (cxSrc + (x - cxDest  + cxSrc) * tcos - (y - cyDest) * tsin) shr 16;
      fy := (cySrc + (x - cxDest) * tsin + (y - cyDest + cySrc) * tcos) shr 16;
      PRGBArray(sDst)[x - x2]:= PRGBArray(sScanLine[fy])[fx];
    end;
    sDst:= sDst + SDstOff;
  end;
  src.Assign(dest);
  dest.free;
  SetLength(sScanLine, 0);
end;

procedure SmoothReSize(Src, Dest: TBitmap);
var
  x, y, px, py: Integer;
  i, x1, x2, z, z2, iz2: Integer;
  w1, w2, w3, w4: Integer;
  Ratio: Integer;
  sDst, sDstOff: Integer;
  sScanLine: Array of PRGBArray;
  Src1, Src2: PRGBArray;
  C, C1, C2: TRGB24;
begin
  sDst:= Integer(src.ScanLine[0]);
  sDstOff:= Integer(src.ScanLine[1]) - Integer(sDst);
  SetLength(sScanLine, Src.Height);
  for i:= 0 to Src.Height - 1 do begin
    sScanLine[i]:= PRGBArray(sDst);
    sDst:= sDst + sDstOff;
  end;
  sDst:= Integer(Dest.ScanLine[0]);
  sDstOff:= Integer(Dest.ScanLine[1]) - sDst;
  Ratio:= ((Src.Width - 1) shl 15) div Dest.Width;
  py:= 0;
  for y := 0 to Dest.Height - 1 do begin
    Src1:= sScanLine[py shr 15];
    if py shr 15 < Src.Height - 1 then Src2:= sScanLine[py shr 15 + 1] else Src2:= Src1;
    z2:= py and $7FFF;
    iz2:= $8000 - z2;
    px:= 0;
    for x := 0 to Dest.Width - 1 do begin
      x1 := px shr 15;
      x2:= x1 + 1;
      C1:= Src1[x1];
      C2:= Src2[x1];
      z:= px and $7FFF;
      w2:= (z * iz2) shr 15;
      w1:= iz2 - w2;
      w4:= (z * z2) shr 15;
      w3:= z2 - w4;
      C.R:= (C1.R * w1 + Src1[x2].R * w2 + C2.R * w3 + Src2[x2].R * w4) shr 15;
      C.G:= (C1.G * w1 + Src1[x2].G * w2 + C2.G * w3 + Src2[x2].G * w4) shr 15;
      C.B:= (C1.B * w1 + Src2[x2].B * w2 + C2.B * w3 + Src2[x2].B * w4) shr 15;
      PRGBArray(sDst)[x]:= C;
      Inc(px, Ratio);
    end;
    sDst:= sDst + SDstOff;
    Inc(py, Ratio);
  end;
  SetLength(sScanLine, 0);
end;

function CalcImgSize(w, h, tw, th: integer): TPoint;
begin
  Result.X := 0;
  Result.Y := 0;
  if (w < tw) and (h < th) then begin
    Result.X := w;
    Result.Y := h;
  end
  else if (w = 0) or (h = 0) then Exit
  else begin
    if w > h then begin
      if w < tw then tw := w;
      Result.X := tw;
      Result.Y := Trunc(tw * h / w);
      if Result.Y > th then begin
        Result.Y := th;
        Result.X := Trunc(th * w / h);
      end;
    end else begin
      if h < th then th := h;
      Result.Y := th;
      Result.X := Trunc(th * w / h);
      if Result.X > tw then begin
        Result.X := tw;
        Result.Y := Trunc(tw * h / w);
      end;
    end;
  end;
end;

procedure TfrmMain.btnSaveClick(Sender: TObject);
begin
  if dlgSave.Execute then begin
    imgSrc.Picture.Bitmap.PixelFormat:= pf24Bit;
    SmoothRotate(imgSrc.Picture.Bitmap, tbAngle.Position / 10);
    imgSrc.Picture.Bitmap.SaveToFile(dlgSave.FileName);
  end;
end;

procedure TfrmMain.chkGridClick(Sender: TObject);
begin
  DoStraighten;
end;

procedure TfrmMain.DoStraighten;
var
  tmp, bmp: TBitmap;
  i: Integer;
  pt: TPoint;
begin
  bmp:= TBitmap.Create;
  bmp.Assign(img);
  bmp.PixelFormat:= pf24Bit;

  if radFast.Checked then
    FastRotate(bmp, tbAngle.Position / 10)
  else
    SmoothRotate(bmp, tbAngle.Position / 10);

  tmp:= TBitmap.Create;
  tmp.PixelFormat:= pf24Bit;
  pt:= CalcImgSize(bmp.Width * 2, bmp.Height * 2, imgDest.Width, imgDest.Height);
  tmp.Width:= pt.X;
  tmp.Height:= pt.Y;
  SmoothReSize(bmp, tmp);

  // PaintGrid?
  if chkGrid.Checked then begin
    tmp.Canvas.Pen.Color:= clWhite;
    tmp.Canvas.Pen.Mode:= pmXor;
    i:= (tmp.Width - ((tmp.Width div 48) * 48)) div 2;
    while i < tmp.Width do begin
      tmp.Canvas.MoveTo(i, 0);
      tmp.Canvas.LineTo(i, tmp.Height - 1);
      i:= i + 48;
    end;
    i:= (tmp.Height - ((tmp.Height div 48) * 48)) div 2;
    while i < tmp.Height do begin
      tmp.Canvas.MoveTo(0, i);
      tmp.Canvas.LineTo(tmp.Width - 1, i);
      i:= i + 48;
    end;
  end;
  imgDest.Picture.Bitmap.Assign(tmp);
  imgDest.Refresh;
  bmp.Free;
  tmp.Free;
end;

procedure TfrmMain.MakePreviewImage;
var
  pt: TPoint;
begin
  if not imgLoaded then Exit;  
  pt:= CalcImgSize(imgSrc.Picture.Width, imgSrc.Picture.Height, imgDest.Width, imgDest.Height);
  img.Width:= pt.X;
  img.Height:= pt.Y;
  imgSrc.Picture.Bitmap.PixelFormat:= pf24Bit;
  SmoothReSize(imgSrc.Picture.Bitmap, img);
  DoStraighten;
end;

procedure TfrmMain.FormCreate(Sender: TObject);
begin
  DoubleBuffered:= True;
  img:= TBitmap.Create;
  img.PixelFormat:= pf24Bit;
  imgLoaded:= False;
end;

procedure TfrmMain.FormDestroy(Sender: TObject);
begin
  img.Free;
end;

procedure TfrmMain.FormResize(Sender: TObject);
begin
  MakePreviewImage;
end;

procedure TfrmMain.tbAngleChange(Sender: TObject);
begin
  labAngle.Caption:= 'Angle: ' + FloatToStrF(tbAngle.Position / 10, ffFixed, 2, 1);
  DoStraighten;
end;

procedure TfrmMain.btnOpenClick(Sender: TObject);
var
  Jpg: TJpegImage;
begin
  if dlgOpen.Execute then begin
    if UpperCase(ExtractFileExt(dlgOpen.FileName)) = '.JPG' then begin
      Jpg:= TJpegImage.Create;
      Jpg.LoadFromFile(dlgOpen.FileName);
      imgSrc.Picture.Bitmap.Width:= Jpg.Width;
      imgSrc.Picture.Bitmap.Height:= Jpg.Height;
      imgSrc.Canvas.Draw(0,0, Jpg);
      Jpg.Free;
    end else imgSrc.Picture.LoadFromFile(dlgOpen.FileName);
    imgLoaded:= True;
  end;
  MakePreviewImage;
end;

end.
